## defaults

# LLM
LLM_PLATFORM        = "openai"
LLM_MODEL           = "gpt-4o-mini"
LLM_EMBEDDING_MODEL = "text-embedding-ada-002"

# Search
LLM_SEARCH_API = "serpapi"

# State Management
LLM_STATE_STORE = "redis"
